# Copyright 2024 Aegiq Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .distribution import Distribution
from .utils import is_number


class Constant(Distribution):
    """
    Implements returns of a singular constant value for the assigned quantity.

    Args:

        value (float) : The constant value to use.

    """

    def __init__(self, value: float) -> None:
        is_number(value)
        self._value = value

    def __str__(self) -> str:
        return f"Constant({self._value})"

    def value(self) -> float:
        """Returns set constant value."""
        return self._value

    def set_random_seed(self, seed: int | None) -> None:
        """Used for setting the random seed for the model."""
